#!/usr/bin/env python

__description__ = 'VBA parsing plugin for oledump.py'
__author__ = 'Didier Stevens'
__version__ = '0.0.2'
__date__ = '2015/09/17'

"""

Source code put in public domain by Didier Stevens, no Copyright
https://DidierStevens.com
Use at your own risk

options:
  -v                  print variable names
  -d var=value        define variable, example: -d ATTH=http
  -u value            define all unknown variables to the same value, example: -u http

History:
  2015/07/15: start
  2015/07/22: detecting obfuscated http: if 'http' in ''.join([c for c in expressionValue.lower() if c.isalpha()])
  2015/08/11: added method calls
  2015/08/18: added stream_joined (for statements split over lines terminated with _)
  2015/08/26: fixed ReplChrFunction regex
  2015/09/04: added option -d
  2015/09/17: added option -u

Todo:
  This is a quick and dirty hack, to be replaced with a BNF parser
"""

import re

def RemoveLeadingWhitespaceAndLineComment(expression):
    expression = expression.lstrip()
    if len(expression) > 0 and expression[0] == "'":
        return ''
    else:
        return expression

def MatchString(expression):
    expression = RemoveLeadingWhitespaceAndLineComment(expression)
    if len(expression) > 0 and expression[0] != '"':
        return None, expression
    else:
        rest = expression[1:]
        position = rest.find('"')
        if position == -1:
            return None, expression
        else:
            return rest[0:position], RemoveLeadingWhitespaceAndLineComment(rest[position + 1:])

def MatchVariable(expression):
    expression = RemoveLeadingWhitespaceAndLineComment(expression)
    if len(expression) > 0 and not expression[0].isalpha():
        return None, expression
    else:
        characters = [c for c in expression]
        variable = ''
        while len(characters) > 0 and (characters[0].isalnum() or characters[0] == '_'):
            variable += characters[0]
            characters = characters[1:]
        return variable, RemoveLeadingWhitespaceAndLineComment(''.join(characters))

def ReplChrFunction(oMatch):
    if re.match(r'^[ 0-9()/*+^-]+$', oMatch.group(1)):
        return '"' + chr(eval(oMatch.group(1))) + '"'
    else:
        return oMatch.group(0)
    
def EvaluateChrFunction(expression):
    return re.sub(r'chrw?\$?\s*\(([^)]+)\)', ReplChrFunction, expression, flags=re.I)

def MatchExpression(expression, dVariables):
    expression = EvaluateChrFunction(expression)
    resultString, restExpression = MatchString(expression)
    if resultString == None:
        variable, restExpression = MatchVariable(expression)
        if variable == None or not variable in dVariables:
            return None
        resultString = dVariables[variable]
    if resultString == None:
        return None
    elif restExpression == '':
        return resultString
    elif len(restExpression) > 0 and restExpression[0] in ('+', '&'):
        restResult = MatchExpression(restExpression[1:], dVariables)
        if restResult == None:
            return None
        return resultString + restResult
    else:
        return None

class cVBA(cPluginParent):
    macroOnly = True
    name = 'VBA parsing plugin'

    def __init__(self, name, stream, options):
        self.streamname = name
        self.stream = stream
        self.stream_joined = stream.replace('_\x0D\x0A', '')
        self.options = options
        self.ran = False

    def ParseVariableDefinitions(self, definitions):
        position = definitions.find('=')
        if position == -1:
            return
        self.dVariables[definitions[0:position].strip()] = definitions[position+1:]
        
    def Analyze(self):
        self.ran = True
        self.dVariables = {}
        self.result = []

        if self.options.startswith('-d'):
            self.ParseVariableDefinitions(self.options[2:])
        	
        # parse assigments
        for oMatch in re.findall(r'^\s*([a-z0-9_]+)\s*=\s*(.+)\s*$', self.stream_joined, re.I + re.M):
            expressionValue = MatchExpression(oMatch[1].strip(), self.dVariables)
            if expressionValue != None:
                if 'http' in ''.join([c for c in expressionValue.lower() if c.isalpha()]):
                    if self.options == '-v':
                        self.result.append(oMatch[0] + ': ' + expressionValue)
                    else:
                        self.result.append(expressionValue)
                self.dVariables[oMatch[0]] = expressionValue
            elif self.options.startswith('-u'):
                self.dVariables[oMatch[0]] = self.options[3:]
        # parse method calls
        for oMatch in re.findall(r'^\s*[a-zA-Z0-9_]+(\.[a-zA-Z0-9_]+)?[ \t]+([^,\n]+([ \t]*,[ \t]*[^,\n]+)*)$', self.stream_joined, re.I + re.M):
            for expression in oMatch[1].split(','):
                expressionValue = MatchExpression(expression.strip(), self.dVariables)
                if expressionValue != None:
                    if 'http' in ''.join([c for c in expressionValue.lower() if c.isalpha()]):
                        self.result.append(expressionValue)
        return self.result

AddPlugin(cVBA)



#!/usr/bin/env python

__description__ = 'Dridex plugin for oledump.py'
__author__ = 'Didier Stevens'
__version__ = '0.0.9'
__date__ = '2015/11/06'

"""

Source code put in public domain by Didier Stevens, no Copyright
https://DidierStevens.com
Use at your own risk

History:
  2015/02/12: start, based on sample 6beaa39b2a1d3d896c5e2fd277c227dd
  2015/02/16: added OlFdL0IOXbF, based on sample f1c80a738722554b91452c59adb2f27d
  2015/02/19: added NewQkeTzIIHM, based on sample d927f8cff07f87c3c3f748604ab35896
  2015/02/25: 0.0.4 added Xor FF, based on sample f3c3fbeed637cccc7549636b7e0f7cdb
  2015/02/26: 0.0.5 added Step2, based on sample 33c5ad38ad766d4e748ee3752fc4c292
  2015/04/08: 0.0.6 added KALLKKKASKAJJAS, based on sample 491A146F5DE3592C7D959E2869F259EF
  2015/04/09: 0.0.7 used KALLKKKASKAJJAS, based on sample 14C2795BCC35C3180649494EC2BC7877
  2015/04/08: 0.0.8 added GQQSfwKSTdAvZbHNhpfK, based on sample 39B38CE4E2E8D843F88C3DF9124527FC
  2015/11/06: 0.0.9 added chr support; added IpkfHKQ2Sd, based on sample 0E73D64FBDF6C87935C0CFF9E65FA3BE

Todo:
"""

import re
import binascii
import array

def RoV(InputStringToBeDecrypted):
    strTempText = InputStringToBeDecrypted
    strText = strTempText
    strDecryptedText = ""
    strText = strText[:len(strText) - 4]
    strText = strText[-(len(strText) - 4):]
    strText, nCharSize = Extract_Char_Size(strText)
    strText, nEncKey = Extract_Enc_Key(strText, nCharSize)
    nTextLenght = len(strText)
    for nCounter in range(0, len(strText), nCharSize):
        strChar1 = strText[nCounter:nCounter + nCharSize]
        nChar = aYP(strChar1)
        nChar2 = nChar / nEncKey
        strChar2 = chr(nChar2)
        strDecryptedText = strDecryptedText + strChar2
    return strDecryptedText.strip()

def Extract_Char_Size(strText):
    nLeft = len(strText) / 2
    strLeft = strText[:nLeft]
    nRight = len(strText) - nLeft
    strRight = strText[-nRight:]
    strKeyEnc = strLeft[-2:]
    strKeySize = strRight[:2]
    strKeyEnc = yiK(strKeyEnc)
    strKeySize = yiK(strKeySize)
    nKeyEnc = int(strKeyEnc)
    nKeySize = int(strKeySize)
    nCharSize = nKeySize - nKeyEnc
    strText = strLeft[:len(strLeft) - 2] + strRight[-(len(strRight) - 2):]
    return (strText, nCharSize)

def yiK(cString):
    strTempString = ""
    for strChar1 in cString:
        if strChar1.isdigit():
            strTempString = strTempString + strChar1
        else:
            strTempString = strTempString + "0"
    return strTempString

def aYP(strTempText):
    strText = ""
    strTempText = strTempText.strip()
    for strChar1 in strTempText:
        if strChar1.isdigit():
            strText = strText + strChar1
    return int(strText)

def Extract_Enc_Key(strText, nCharSize):
    strEncKey = ""
    nLenght = len(strText) - nCharSize
    nLeft = nLenght / 2
    strLeft = strText[:nLeft]
    nRight = nLenght - nLeft
    strRight = strText[-nRight:]
    strEncKey = strText[nLeft:nLeft + nCharSize]
    strEncKey = yiK(strEncKey)
    nEncKey = int(strEncKey.strip())
    strText = strLeft + strRight
    return (strText, nEncKey)

def MakePositive(value1, value2):
    while value1 < 0:
        value1 += value2
    return value1

def OlFdL0IOXbF(InputData, NumKey):
    return ''.join([chr(MakePositive(ord(c), 256) - NumKey) for c in InputData])

def NewQkeTzIIHM(InputData):
    return ''.join([chr(ord(c) - 13) for c in InputData])

def lqjWjFO(strData, strKey):
    result = ''
    for iIter in range(len(strData)):
        if iIter < len(strKey):
            result += chr(ord(strData[iIter]) - ord(strKey[iIter]))
        else:
            result += chr(ord(strData[iIter]) - ord(strKey[iIter % (len(strKey) - 1)]))
    return result

def Xor(data, key):
    return ''.join([chr(ord(c) ^ key) for c in data])

def Step(data, step):
    result = ''
    for iIter in range(0, len(data), step):
        result += data[iIter]
    return result

def ContainsString(listStrings, key):
    for aString in listStrings:
        if key.lower() in aString.lower():
            return True
    return False

def IsHex(value):
    return re.match(r'^([0-9a-f][0-9a-f])+$', value, re.IGNORECASE) != None

def KALLKKKASKAJJAS(strKey, strData):
    result = ''
    encoded = binascii.a2b_hex(strData)
    for iIter in range(len(encoded)):
        result += chr(ord(encoded[iIter]) ^ ord(strKey[(((iIter + 1) % len(strKey)))]))
    return result

def GQQSfwKSTdAvZbHNhpfK(strData, strKey):
    result = ''
    dX = {x:0 for x in range(256)}
    Y = 0
    for iIter in range(256):
        Y = (Y + dX[iIter] + ord(strKey[iIter % len(strKey)])) % 256
        dX[iIter] = iIter
    for iIter in range(len(strData)):
        Y = (Y + dX[Y] + 1) % 256
        result += chr(ord(strData[iIter]) ^ dX[dX[(Y + dX[Y]) % 254]])

    return result

def IpkfHKQ2Sd(secret, key):
    aTable = array.array('i', [0] * (285 + 1))
    aSecret = array.array('i', [0] * len(secret))
    keyLength = len(key) - 1
    for iIter in range(0, 255 + 1):
        aTable[iIter] = iIter
    for iIter in range(256, 285 + 1):
        aTable[iIter] = iIter ^ 256
    for iIter in range(1, 6 + 1):
        aTable[iIter + 249] = ord(key[keyLength - iIter])
        aTable[iIter - 1] = ord(key[iIter - 1]) ^ (255 - ord(key[keyLength - iIter]))

    bCondition = False
    indexKey = 0
    indexTable = 0
    for iIter in range(0, len(secret) - 1 + 1):
        if indexKey > keyLength:
            indexKey = 0
        if indexTable > 285 and bCondition == False:
            indexTable = 0
            bCondition = not bCondition
        if indexTable > 285 and bCondition == True:
            indexTable = 5
            bCondition = not bCondition
        aSecret[iIter] = ord(secret[iIter]) ^ (aTable[indexTable] ^ ord(key[indexKey]))
        indexKey = indexKey + 1
        indexTable = indexTable + 1
    return ''.join(map(chr, aSecret))

class cDridexDecoder(cPluginParent):
    macroOnly = True
    name = 'Dridex decoder'

    def __init__(self, name, stream, options):
        self.streamname = name
        self.stream = stream
        self.options = options
        self.ran = False

    def Analyze(self):
        self.ran = True

        oREString = re.compile(r'"([^"\n]+)"')
        foundStrings = oREString.findall(self.stream)
        oREChr = re.compile(r'((chr[w\$]?\(\d+\)(\s*[&+]\s*)?)+)', re.IGNORECASE)
        oREDigits = re.compile(r'\d+')
        for foundTuple in oREChr.findall(self.stream.replace('_\r\n', '')):
            chrString = ''.join(map(lambda x: chr(int(x)), oREDigits.findall(foundTuple[0])))
            if chrString != '':
                foundStrings.append(chrString)

        for DecodingFunction in [RoV, lambda s:OlFdL0IOXbF(s, 61), NewQkeTzIIHM, lambda s:Xor(s, 0xFF), lambda s:Step(s, 2)]:
            result = []
            for foundString in foundStrings:
                try:
                    result.append(DecodingFunction(foundString))
                except:
                    pass

            if ContainsString(result, 'http'):
                return result

        foundStringsSmall = [foundString for foundString in foundStrings if len(foundString) <= 10]
        foundStringsLarge = [foundString for foundString in foundStrings if len(foundString) > 10]
        for foundStringSmall in foundStringsSmall:
            for DecodingFunction in [lqjWjFO, GQQSfwKSTdAvZbHNhpfK, IpkfHKQ2Sd]:
                result = []
                for foundStringLarge in foundStringsLarge:
                    try:
                        result.append(DecodingFunction(foundStringLarge, foundStringSmall))
                    except:
                        pass

                if ContainsString(result, 'http:'):
                    return result

        foundStringsHex = [foundString for foundString in foundStrings if IsHex(foundString)]
        foundStringsNotHex = [foundString for foundString in foundStrings if not IsHex(foundString)]
        for foundStringNotHex in foundStringsNotHex:
            for DecodingFunction in [KALLKKKASKAJJAS]:
                result = []
                for foundStringHex in foundStringsHex:
                    try:
                        result.append(DecodingFunction(foundStringNotHex, foundStringHex))
                    except:
                        pass

                if ContainsString(result, 'http'):
                    return result

        for foundStringHex1 in foundStringsHex:
            for DecodingFunction in [KALLKKKASKAJJAS]:
                result = []
                for foundStringHex2 in foundStringsHex:
                    try:
                        result.append(DecodingFunction(foundStringHex1, foundStringHex2))
                    except:
                        pass

                if ContainsString(result, 'http'):
                    return result

        return []

AddPlugin(cDridexDecoder)



#!/usr/bin/env python

__description__ = 'jumplist plugin for oledump.py'
__author__ = 'Didier Stevens'
__version__ = '0.0.1'
__date__ = '2015/05/12'

"""

Source code put in public domain by Didier Stevens, no Copyright
https://DidierStevens.com
Use at your own risk

History:
  2015/04/14: start
  2015/05/11: added option -f
  2015/05/12: continue

Todo:
"""

import struct
import datetime
import binascii
import uuid

class cJumpList(cPluginParent):
    macroOnly = False
    name = 'jumplist plugin'

    def __init__(self, name, stream, options):
        self.streamname = name
        self.stream = stream
        self.options = options
        self.ran = False

# http://articles.forensicfocus.com/2012/10/30/forensic-analysis-of-windows-7-jump-lists/
    def Analyze(self):
        result = []
        fullOutput = self.options != '-f'
        if self.streamname == ['DestList']:
            self.ran = True
            stream = self.stream
            headerFormat = 'IIIfQQ'
            formatsize = struct.calcsize(headerFormat)
            firstIssuedEntryID, totalNumberCurrentEntries, totalNumberPinnedEntries, float, lastIssuedEntryID, numberOfActions = struct.unpack(headerFormat, stream[:formatsize])
            stream = stream[formatsize:]
            if fullOutput:
                result.append('First Issued Entry ID: %d' % firstIssuedEntryID)
                result.append('Total number of current entries in Jump List: %d' % totalNumberCurrentEntries)
                result.append('Total number of pinned entries: %d' % totalNumberPinnedEntries)
                result.append('Floating point value. Some kind of counter: %f' % float)
                result.append('Last issued Entry ID number: %d' % lastIssuedEntryID)
                result.append('Number of add/delete actions: %d' % numberOfActions)
                result.append('')
                result.append('Entries:')
            while len(stream) > 114:
                if fullOutput:
                    checksum = struct.unpack('Q', stream[0:8])[0]
                    newVolumeID = stream[8:24]
                    objectID1 = stream[24:40]
                    birthVolumeID = stream[40:56]
                    objectID2 = stream[56:72]
                    netbiosName = stream[72:88]
                    entryID = struct.unpack('Q', stream[88:96])[0]
                    floatAccess = struct.unpack('f', stream[96:100])[0]
                    msfiletime = struct.unpack('Q', stream[100:108])[0]
                    entryPinStatus = struct.unpack('I', stream[108:112])[0]

                    result.append('')
                    result.append(' Checksum: %x' % checksum)
                    result.append(' New volume ID: %s' % binascii.b2a_hex(newVolumeID))
                    result.append(' New volume ID: %s' % uuid.UUID(bytes=newVolumeID))
                    result.append(' Object ID 1: %s' % binascii.b2a_hex(objectID1))
                    result.append(' Birth volume ID: %s' % binascii.b2a_hex(birthVolumeID))
                    result.append(' Object ID 2: %s' % binascii.b2a_hex(objectID2))
                    result.append(' Netbios Name: %s' % netbiosName.rstrip('\0'))
                    result.append(' Entry ID number: %d' % entryID)
                    result.append(' Floating point access counter: %f' % floatAccess)
                    result.append(' Filetime: ' + (datetime.datetime(1601, 1, 1, 0, 0, 0) + datetime.timedelta(microseconds=msfiletime/10)).isoformat())
                    result.append(' Entry Pin status: %d' % entryPinStatus)
                stream = stream[112:]
                size = struct.unpack('H', stream[:2])[0]
                stream = stream[2:]
                if fullOutput:
                    result.append(' Filename: ' + stream[:size * 2].decode('utf-16'))
                else:
                    result.append(stream[:size * 2].decode('utf-16'))
                stream = stream[size * 2:]
            if len(stream) > 0:
                if fullOutput:
                    result.append('')
                    result.append('Remaining bytes: %d' % len(stream))

        return result

AddPlugin(cJumpList)



#!/usr/bin/env python

__description__ = 'String subtract plugin for oledump.py'
__author__ = 'Didier Stevens'
__version__ = '0.0.1'
__date__ = '2017/03/04'

"""

Source code put in public domain by Didier Stevens, no Copyright
https://DidierStevens.com
Use at your own risk

History:
  2017/03/04: start

Todo:
"""

import re

def AllCharactersInStr(str1, str2):
    return all([c in str2 for c in str1])

class cStringSubtract(cPluginParent):
    macroOnly = True
    name = 'String subtract plugin'

    def __init__(self, name, stream, options):
        self.streamname = name
        self.stream = stream
        self.options = options
        self.ran = False

    def Analyze(self):
        self.ran = True

        allstrings = []
        result = []
        dDecodes = {}

        oREStr = re.compile(r'"[^"\n]+"')
        for found in oREStr.findall(self.stream.replace('_\r\n', '')):
            allstrings.append(found)
        for str1 in allstrings:
            for str2 in allstrings:
                if str1 != str2 and AllCharactersInStr(str1, str2):
                    if not str1 in dDecodes:
                        dDecodes[str1] = []
                    dDecodes[str1].append(''.join([c for c in str2 if not c in str1]))
                    
        for v, k in sorted([(len(v), k) for k, v in dDecodes.items()]):
            for s in dDecodes[k]:
                result.append(s)
        return result

AddPlugin(cStringSubtract)



#!/usr/bin/env python

__description__ = 'Office crypto plugin for oledump.py'
__author__ = 'Didier Stevens'
__version__ = '0.0.1'
__date__ = '2018/05/06'

"""

Source code put in public domain by Didier Stevens, no Copyright
https://DidierStevens.com
Use at your own risk

History:
  2018/05/06: start

Todo:
"""

import struct

class cOfficeCrypto(cPluginParent):
    macroOnly = False
    name = 'Office crypto plugin'

    def __init__(self, name, stream, options):
        self.streamname = name
        self.stream = stream
        self.options = options
        self.ran = False

    def Analyze(self):
        result = []

        if self.streamname == ['EncryptionInfo']:
            self.ran = True
            if len(self.stream) >= 4:
                dVersions = {'2.2': 'Standard Encryption', '3.2': 'Standard Encryption', '4.2': 'Standard Encryption', '3.3': 'Extensible Encryption', '4.3': 'Extensible Encryption', '4.4': 'Agile Encryption'}
                versionMajor, versionMinor = struct.unpack("<HH", self.stream[0:4])
                version = '%d.%d' % (versionMajor, versionMinor)
                result.append('Crypto version %s: %s' % (version, dVersions.get(version, 'Unknown')))
            else:
                result.append('EncryptionInfo stream is too short')

        return result

AddPlugin(cOfficeCrypto)



#!/usr/bin/env python

__description__ = 'Word DTTM date/time structure plugin for oledump.py'
__author__ = 'Didier Stevens'
__version__ = '0.0.1'
__date__ = '2022/11/09'

"""

Source code put in public domain by Didier Stevens, no Copyright
https://DidierStevens.com
Use at your own risk

History:
  2022/10/09: start
  2022/11/09: added option --verbose

Todo:
"""

import bitstruct
import datetime
import collections

#https://interoperability.blob.core.windows.net/files/MS-DOC/%5bMS-DOC%5d.pdf

nDTTM = collections.namedtuple('nDTTM', ['valid', 'null', 'year', 'month', 'day', 'weekday', 'hours', 'minutes'])

def ParseDTTM(data):
    if data == b'\x00\x00\x00\x00':
        return nDTTM(True, True, 0, 0, 0, 0, 0, 0)
    inDTTM = nDTTM(False, False, 0, 0, 0, 0, 0, 0)
    weekday, year, month, day, hours, minutes = bitstruct.unpack('u3u9u4u5u5u6', data[::-1])
    if minutes > 0x3B or hours > 0x17 or day == 0 or day > 0x1F or month == 0 or month > 0x0C or weekday > 0x06:
        return inDTTM
    year += 1900
    try:
        oDatetime = datetime.datetime(year, month, day)
    except ValueError:
        return inDTTM
    weekdayCheck = oDatetime.weekday() + 1
    if weekdayCheck == 7:
        weekdayCheck = 0
    if weekdayCheck != weekday:
        return inDTTM
    return nDTTM(True, False, year, month, day, weekday, hours, minutes)

def PrintDTTM(inDTTM):
    if inDTTM.null:
        return 'null'
    dWeekdays = {
        0: 'Sun',
        1: 'Mon',
        2: 'Tue',
        3: 'Wed',
        4: 'Thu',
        5: 'Fri',
        6: 'Sat',
    }
    return '%04d/%02d/%02d(%s) %02d:%02d' % (inDTTM.year, inDTTM.month, inDTTM.day, dWeekdays[inDTTM.weekday], inDTTM.hours, inDTTM.minutes)

class cCLSID(cPluginParent):
    macroOnly = False
    name = 'OLE streams plugin'

    def __init__(self, name, stream, options):
        self.streamname = name
        self.stream = stream
        self.options = options
        self.ran = False

    def Analyze(self):
        oParser = optparse.OptionParser()
        oParser.add_option('--minyear', type=int, default=1900, help='Minimum year value (default 1900)')
        oParser.add_option('--maxyear', type=int, default=3000, help='Maximum year value (default 3000)')
        oParser.add_option('-V', '--verbose', action='store_true', default=False, help='verbose output')
        (options, args) = oParser.parse_args(self.options.split(' '))

        result = []
        self.ran = True
        stream = self.stream
        positionPotentialDOPs = []
        potentialPrevious = []

        dPrevious = {
            0x6805: 'sprmCDttmRMark',
            0x6864: 'sprmCDttmRMarkDel',
        }

        for iter in range(len(stream) - 4):
            data = self.stream[iter:iter + 4]
            inDTTM = ParseDTTM(data)
            if not inDTTM.valid:
                continue
            if inDTTM.null:
                continue
            if inDTTM.year < options.minyear or inDTTM.year > options.maxyear:
                continue
            if options.verbose:
                result.append('0x%08x: %s' % (iter, PrintDTTM(inDTTM)))
            if iter >= 2:
                previous = struct.unpack('<H', self.stream[iter - 2:iter])[0]
                if previous == 0x0000:
                    positionPotentialDOPs.append(iter)
                if previous in dPrevious:
                    potentialPrevious.append([previous, iter])

        validDOPs = []
        for position in positionPotentialDOPs:
            format = '<HIIIHI'
            lengthData = 3*4 + struct.calcsize(format)
            data = stream[position:position+lengthData]
            if len(data) != lengthData:
                continue
            dataCreated = data[:4]
            dataRevised = data[4:8]
            dataLastPrint = data[8:12]
            inDTTMCreated = ParseDTTM(dataCreated)
            inDTTMRevised = ParseDTTM(dataRevised)
            inDTTMLastPrint = ParseDTTM(dataLastPrint)
            if inDTTMCreated.valid and (inDTTMRevised.valid or inDTTMRevised.null) and (inDTTMLastPrint.valid or inDTTMLastPrint.null):
                validDOPs.append('Position DOP: 0x%08x' % position)
                validDOPs.append(' dttmCreated:   %s %s' % (PrintDTTM(inDTTMCreated), binascii.b2a_hex(dataCreated)))
                validDOPs.append(' dttmRevised:   %s %s' % (PrintDTTM(inDTTMRevised), binascii.b2a_hex(dataRevised)))
                validDOPs.append(' dttmLastPrint: %s %s' % (PrintDTTM(inDTTMLastPrint), binascii.b2a_hex(dataLastPrint)))
                metadata = struct.unpack(format, data[12:])
                validDOPs.append(' nRevision: %d' % metadata[0])
                validDOPs.append(' tmEdited: %d' % metadata[1])
                validDOPs.append(' cWords: %d' % metadata[2])
                validDOPs.append(' cCh: %d' % metadata[3])
                validDOPs.append(' cPg: %d' % metadata[4])
                validDOPs.append(' cParas: %d' % metadata[5])
        result.extend(validDOPs)

        previousOutput = []
        for previous, position in potentialPrevious:
            dataDTTM = stream[position:position+4]
            previousOutput.append(' %s: %s' % (dPrevious.get(previous, '0x%04x' % previous), PrintDTTM(ParseDTTM(dataDTTM))))

        if len(previousOutput) > 0:
            result.append('DTTMs:')
            result.extend(previousOutput)

        return result

AddPlugin(cCLSID)



#!/usr/bin/env python

__description__ = 'HYPERLINK plugin for oledump.py'
__author__ = 'Didier Stevens'
__version__ = '0.0.1'
__date__ = '2025/05/04'

"""

Source code put in public domain by Didier Stevens, no Copyright
https://DidierStevens.com
Use at your own risk

History:
  2025/05/04: start

Todo:
"""

class cHYPERLINK(cPluginParentOle):
    name = 'HYPERLINK plugin'

    def PreProcess(self):
        self.urls = set()

    def Process(self, name, stream):
        for url in re.findall(b'HYPERLINK "(.+?)"', stream):
            if not url in self.urls:
                print(url.decode())
                self.urls.add(url)

    def PostProcess(self):
        pass

AddPlugin(cHYPERLINK)



#!/usr/bin/env python

__description__ = 'OLE streams plugin for oledump.py'
__author__ = 'Didier Stevens'
__version__ = '0.0.2'
__date__ = '2022/06/13'

"""

Source code put in public domain by Didier Stevens, no Copyright
https://DidierStevens.com
Use at your own risk

History:
  2022/06/06: start
  2022/06/06: 0.0.2 continue
  2022/06/08: continue
  2022/06/13: continue

Todo:
"""

duriFlags = {
    'A': 'createAllowRelative (1 bit): A bit that specifies that if the URI scheme is unspecified and not implicitly "file," a relative scheme is assumed during creation of the URI.',
    'B': 'createAllowImplicitWildcardScheme (1 bit): A bit that specifies that if the URI scheme is unspecified and not implicitly "file," a wildcard scheme is assumed during creation of the URI.',
    'C': 'createAllowImplicitFileScheme (1 bit): A bit that specifies that if the URI scheme is unspecified and the URI begins with a drive letter or a UNC path, a file scheme is assumed during creation of the URI.',
    'D': 'createNoFrag (1 bit): A bit that specifies that if a URI query string is present, the URI fragment is not looked for during creation of the URI.',
    'E': 'createNoCanonicalize (1 bit): A bit that specifies that the scheme, host, authority, path, and fragment will not be canonicalized during creation of the URI. This value MUST be 0 if createCanonicalize equals 1.',
    'F': 'createCanonicalize (1 bit): A bit that specifies that the scheme, host, authority, path, and fragment will be canonicalized during creation of the URI. This value MUST be 0 if createNoCanonicalize equals 1.',
    'G': 'createFileUseDosPath (1 bit): A bit that specifies that MS-DOS path compatibility mode will be used during creation of file URIs.',
    'H': 'createDecodeExtraInfo (1 bit): A bit that specifies that percent encoding and percent decoding canonicalizations will be performed on the URI query and URI fragment during creation of the URI. This field takes precedence over the createNoCanonicalize field. This value MUST be 0 if createNoDecodeExtraInfo equals 1. The value 1 can also be saved. This will cause a return value of E_INVALIDARG from CreateUri().',
    'I': 'createNoDecodeExtraInfo (1 bit): A bit that specifies that percent encoding and percent decoding canonicalizations will not be performed on the URI query and URI fragment during creation of the URI. This field takes precedence over the createCanonicalize field. This value MUST be 0 if createDecodeExtraInfo equals 1. The value 1 can also be saved. This will cause a return value of E_INVALIDARG from CreateUri().',
    'J': 'createCrackUnknownSchemes (1 bit): A bit that specifies that hierarchical URIs with unrecognized URI schemes will be treated like hierarchical URIs during creation of the URI. This value MUST be 0 if createNoCrackUnknownSchemes equals 1.',
    'K': 'createNoCrackUnknownSchemes (1 bit): A bit that specifies that hierarchical URIs with unrecognized URI schemes will be treated like opaque URIs during creation of the URI. This value MUST be 0 if createCrackUnknownSchemes equals 1.',
    'L': 'createPreProcessHtmlUri (1 bit): A bit that specifies that preprocessing will be performed on the URI to remove control characters and white space during creation of the URI. This value MUST be 0 if createNoPreProcessHtmlUri equals 1.',
    'M': 'createNoPreProcessHtmlUri (1 bit): A bit that specifies that preprocessing will not be performed on the URI to remove control characters and white space during creation of the URI. This value MUST be 0 if createPreProcessHtmlUri equals 1.',
    'N': 'createIESettings (1 bit): A bit that specifies that registry settings will be used to determine default URL parsing behavior during creation of the URI. This value MUST be 0 if createNoIESettings equals 1.',
    'O': 'createNoIESettings (1 bit): A bit that specifies that registry settings will not be used to determine default URL parsing behavior during creation of the URI. This value MUST be 0 if createIESettings equals 1.',
    'P': 'createNoEncodeForbiddenCharacters (1 bit): A bit that specifies that URI characters forbidden in [RFC3986] will not be percent-encoded during creation of the URI.',
}

def GUIDToBytes(clsid):
    parts = [binascii.a2b_hex(part) for part in clsid.split('-')]
    return parts[0][::-1] + parts[1][::-1] + parts[2][::-1] + parts[3] + parts[4]

def BytesToGUID(data):
    return '{%02X%02X%02X%02X-%02X%02X-%02X%02X-%02X%02X-%02X%02X%02X%02X%02X%02X}' % tuple(data[3::-1] + data[5:3:-1] + data[7:5:-1] + data[8:16])


def Unpack(format, data):
    size = struct.calcsize(format)
    if size > len(data):
        return []
    result = list(struct.unpack(format, data[:size]))
    result.append(data[size:])
    return result

def UnpackCounterBytes(format, data):
    size = struct.calcsize(format)
    if size > len(data):
        return []
    length = struct.unpack(format, data[:size])[0]
    data = data[size:]
    if length == 0:
        result = [b'']
    elif length > len(data):
        return []
    else:
        result = [data[:length]]

    result.append(data[length:])
    return result

def UnpackCounterBytes2(format, data):
    size = struct.calcsize(format)
    if size > len(data):
        return []
    length = struct.unpack(format, data[:size])[0] * 2
    data = data[size:]
    if length == 0:
        result = [b'']
    elif length > len(data):
        return []
    else:
        result = [data[:length]]

    result.append(data[length:])
    return result

def UnpackStream(format, data):
    size = struct.calcsize(format)
    if size > len(data):
        return []
    length = struct.unpack(format, data[:size])[0]
    data = data[size:]
    if length == 0:
        result = [b'']
    else:
        length -= size
        if length > len(data):
            return []
        else:
            result = [data[:length]]

    result.append(data[length:])
    return result

def UnpackBytes(size, data):
    if size > len(data):
        return []
    return [data[:size], data[size:]]

def UnpackNULLTerminatedUNICODEString(data):
    string = b''
    while len(data) >= 2 and data[:2] != b'\x00\x00':
        string += data[:2]
        data = data[2:]
    return [string.decode('utf16'), data[2:]]

def Indent(indentation, list):
    return [indentation + item for item in list]

def InterpretBits(value, dBits, falseToo=False):
    result = []

    for key in dBits.keys():
        if isinstance(key, str):
            position = ord(key) - ord('A')
        else:
            position = key
        position = 2 ** position
        if value & position == position:
            result.append('1: %s %s' % (key, dBits[key]))
        elif falseToo:
            result.append('0: %s %s' % (key, dBits[key]))

    return result

#https://docs.microsoft.com/en-us/windows/win32/api/objbase/nf-objbase-createitemmoniker
def ParseItemMonikerStream(data):
    returnvalue = UnpackCounterBytes('<I', data)
    if returnvalue == []:
        return []
    delimiter, data = returnvalue
    returnvalue = UnpackCounterBytes('<I', data)
    if returnvalue == []:
        return []
    item, data = returnvalue
    if returnvalue == []:
        return []
    result = [delimiter.decode('utf16'), item.decode('latin').rstrip('\x00'), data]
    return result

#URLMoniker 2.3.7.6 https://interoperability.blob.core.windows.net/files/MS-OSHARED/%5bMS-OSHARED%5d.pdf
def ParseURLMonikerStream(data):
    returnvalue = Unpack('<I', data)
    if returnvalue == []:
        return []
    length, data = returnvalue
    dataURLMoniker = data[:length]
    returnvalue = UnpackNULLTerminatedUNICODEString(dataURLMoniker)
    if returnvalue == []:
        url = '<ERROR>'
        remainder = b''
    else:
        url, remainder = returnvalue

    returnvalue = UnpackBytes(16, remainder)
    if returnvalue == []:
        clsid = '<ERROR>'
        remainder = b''
    else:
        clsid, remainder = returnvalue
        clsid = BytesToGUID(clsid)

    returnvalue = Unpack('<I', remainder)
    if returnvalue == []:
        version = '<ERROR>'
        remainder = b''
    else:
        version, remainder = returnvalue

    returnvalue = Unpack('<I', remainder)
    if returnvalue == []:
        uriFlags = '<ERROR>'
        remainder = b''
    else:
        uriFlags, remainder = returnvalue

    return [url, clsid, version, uriFlags, data[length:]]

#https://interoperability.blob.core.windows.net/files/MS-OSHARED/%5bMS-OSHARED%5d.pdf
def ParseCompositeMonikerStream(data):
    returnvalue = Unpack('<I', data)
    if returnvalue == []:
        return []
    counter, data = returnvalue

    monikerResults = []
    while len(data) > 0:
        if len(data) < 16:
            return []
        clsid = BytesToGUID(data[:16])
        data = data[16:]
        description = 'clsid: %s (%s)' % (clsid, LookupClsID(clsid))
        if clsid == '{79EAC9E0-BAF9-11CE-8C82-00AA004BA90B}':
            returnvalue = ParseURLMonikerStream(data)
            if returnvalue == []:
                return []
            if returnvalue[1] == '{F4815879-1D3B-487F-AF2C-825DC4852763}':
                serialGUIDExpected = ' (expected value)'
            else:
                serialGUIDExpected = ' (different from expected value {F4815879-1D3B-487F-AF2C-825DC4852763})'
            result = [description, 'URL: %s' % returnvalue[0], 'SerialGUID: %s%s' % (returnvalue[1], serialGUIDExpected), 'Version: %s' % returnvalue[2], 'uriFlags: %s' % returnvalue[3]]
            if isinstance(returnvalue[3], int):
                result.extend(Indent('  ', InterpretBits(returnvalue[3], duriFlags)))
            monikerResults.append(result)
            data = returnvalue[4]
        elif clsid == '{00000304-0000-0000-C000-000000000046}':
            returnvalue = ParseItemMonikerStream(data)
            if returnvalue == []:
                return []
            result = [description, 'Delimiter: %s' % returnvalue[0], 'Item: %s' % returnvalue[1]]
            monikerResults.append(result)
            data = returnvalue[2]
        else:
            raise

    result = [counter, monikerResults, data]
    return result

def ParseMonikerStream(data):
    result = []
    if len(data) < 16:
        return []
    clsid = BytesToGUID(data[:16])
    data = data[16:]
    if clsid == '{00000304-0000-0000-C000-000000000046}':
        returnvalue = ParseItemMonikerStream(data)
        if returnvalue != []:
            result = ['Delimiter: %s' % returnvalue[0], 'Item: %s' % returnvalue[1]]
            data = returnvalue[2]
    elif clsid == '{00000309-0000-0000-C000-000000000046}':
        returnvalue = ParseCompositeMonikerStream(data)
        if returnvalue != []:
            result = ['Counter: %d' % returnvalue[0]]
            for items in returnvalue[1]:
                result.append('  ' + items[0])
                result.extend(Indent('    ', items[1:]))
            data = returnvalue[2]
    elif clsid == '{79EAC9E0-BAF9-11CE-8C82-00AA004BA90B}':
        returnvalue = ParseURLMonikerStream(data)
        if returnvalue != []:
            result = ['URL: %s' % returnvalue[0], 'SerialGUID: %s' % returnvalue[1], 'Version: %s' % returnvalue[2], 'uriFlags: %s' % returnvalue[3]]
            if isinstance(returnvalue[3], int):
                result.extend(Indent('  ', InterpretBits(returnvalue[3], duriFlags)))
            data = returnvalue[4]
    return [clsid, result, data]

def LookupClsID(clsid):
    clsidNormalized = clsid.strip('{').strip('}').upper()
    return KNOWN_CLSIDS.get(clsidNormalized, '<UNKNOWN>')

def FindCLSIDs(data):
    result = []
    if KNOWN_CLSIDS == {}:
        result.append('<oletools missing>')
    for clsid, desc in KNOWN_CLSIDS.items():
        for position in FindAll(data, GUIDToBytes(clsid)):
            result.append('0x%08x %s %s' % (position, clsid, desc))

    return sorted(result)

def UnpackClipboardFormatOrString(data):
    returnvalue = Unpack('<I', data)
    if returnvalue == []:
        return []
    markerOrLength, dataRemainder = returnvalue
    if markerOrLength == 0:
        return [markerOrLength, None, dataRemainder]
    if markerOrLength in [0xFFFFFFFE, 0xFFFFFFFF]:
        returnvalue = Unpack('<I', dataRemainder)
        if returnvalue == []:
            return []
        clipboardFormat, dataRemainder = returnvalue
        return [markerOrLength, clipboardFormat, dataRemainder]
    else:
        returnvalue = UnpackBytes(markerOrLength, dataRemainder)
        if returnvalue == []:
            return []
        ansiString, dataRemainder = returnvalue
        return [markerOrLength, ansiString, dataRemainder]

class cCLSID(cPluginParent):
    macroOnly = False
    name = 'OLE streams plugin'

    def __init__(self, name, stream, options):
        self.streamname = name
        self.stream = stream
        self.options = options
        self.ran = False

    def Analyze1Ole(self):
        result = []

        returnvalue = Unpack('<I', self.stream)
        if returnvalue == []:
            return result
        version, streamRemainder = returnvalue
        result.append('Version: %04x' % version)

        returnvalue = Unpack('<I', streamRemainder)
        if returnvalue == []:
            return result
        flags, streamRemainder = returnvalue
        result.append('Flags: %s' % bin(flags))
        if 0x01 & flags == 0x01:
            result.append('  Linked object')
        else:
            result.append('  Embedded object')
        if 0x08 & flags == 0x08:
            result.append('  Implementation specific bit set')

        returnvalue = Unpack('<I', streamRemainder)
        if returnvalue == []:
            return result
        linkUpdateOption, streamRemainder = returnvalue
        result.append('LinkUpdateOption: %04x' % linkUpdateOption)

        returnvalue = Unpack('<I', streamRemainder)
        if returnvalue == []:
            return result
        reserved1, streamRemainder = returnvalue
        result.append('Reserved1: %04x' % reserved1)

        returnvalue = UnpackStream('<I', streamRemainder)
        if returnvalue == []:
            return result
        reservedMonikerStream, streamRemainder = returnvalue
        if len(reservedMonikerStream) > 0:
            result.append('ReservedMonikerStream present')
            returnvalue = ParseMonikerStream(reservedMonikerStream)
            if returnvalue == []:
                return result
            clsidMonikerstream, resultMoniker, remainderMoniker = returnvalue
            result.append('  clsid: %s (%s)' % (clsidMonikerstream, LookupClsID(clsidMonikerstream)))
            result.extend(Indent('  ', resultMoniker))
            if len(remainderMoniker) > 0:
                result.append('  extra bytes: %d' % len(remainderMoniker))
        else:
            result.append('ReservedMonikerStream not present')

        returnvalue = UnpackStream('<I', streamRemainder)
        if returnvalue == []:
            return result
        relativeSourceMonikerStream, streamRemainder = returnvalue
        if len(relativeSourceMonikerStream) > 0:
            result.append('RelativeSourceMonikerStream present')
            returnvalue = ParseMonikerStream(relativeSourceMonikerStream)
            if returnvalue == []:
                return result
            clsidMonikerstream, resultMoniker, remainderMoniker = returnvalue
            result.append('  clsid: %s (%s)' % (clsidMonikerstream, LookupClsID(clsidMonikerstream)))
            result.extend(Indent('  ', resultMoniker))
            if len(remainderMoniker) > 0:
                result.append('  extra bytes: %d' % len(remainderMoniker))
        else:
            result.append('RelativeSourceMonikerStream not present')

        returnvalue = UnpackStream('<I', streamRemainder)
        if returnvalue == []:
            return result
        absoluteSourceMonikerStream, streamRemainder = returnvalue
        if len(absoluteSourceMonikerStream) > 0:
            result.append('AbsoluteSourceMonikerStream present')
            returnvalue = ParseMonikerStream(absoluteSourceMonikerStream)
            if returnvalue == []:
                return result
            clsidMonikerstream, resultMoniker, remainderMoniker = returnvalue
            result.append('  clsid: %s (%s)' % (clsidMonikerstream, LookupClsID(clsidMonikerstream)))
            result.extend(Indent('  ', resultMoniker))
            if len(remainderMoniker) > 0:
                result.append('  extra bytes: %d' % len(remainderMoniker))
                result.append(repr(remainderMoniker))
        else:
            result.append('AbsoluteSourceMonikerStream not present')

        returnvalue = Unpack('<I', streamRemainder)
        if returnvalue == []:
            return result
        clsidIndicator, streamRemainder = returnvalue
        result.append('ClsidIndicator: %04x' % clsidIndicator)

        returnvalue = UnpackBytes(16, streamRemainder)
        if returnvalue == []:
            return result
        clsid, streamRemainder = returnvalue
        result.append('Clsid: %s' % BytesToGUID(clsid))

        returnvalue = UnpackCounterBytes('<I', streamRemainder)
        if returnvalue == []:
            return result
        reservedDisplayName, streamRemainder = returnvalue
        result.append('ReservedDisplayName: %s' % reservedDisplayName.decode('utf16').rstrip('\x00'))

        returnvalue = Unpack('<I', streamRemainder)
        if returnvalue == []:
            return result
        reserved2, streamRemainder = returnvalue
        result.append('Reserved2: %08x' % reserved2)

        returnvalue = Unpack('<Q', streamRemainder)
        if returnvalue == []:
            return result
        localUpdateTime, streamRemainder = returnvalue
        result.append('LocalUpdateTime: %016x' % localUpdateTime)

        returnvalue = Unpack('<Q', streamRemainder)
        if returnvalue == []:
            return result
        localCheckUpdateTime, streamRemainder = returnvalue
        result.append('LocalCheckUpdateTime: %016x' % localCheckUpdateTime)

        returnvalue = Unpack('<Q', streamRemainder)
        if returnvalue == []:
            return result
        remoteUpdateTime, streamRemainder = returnvalue
        result.append('RemoteUpdateTime: %016x' % remoteUpdateTime)

        if len(streamRemainder) > 0:
            result.append('extra bytes: %d' % len(streamRemainder))

        return result

    #have yet to find documentation for \003LinkInfo
    def Analyze3LinkInfo(self):
        result = []

        returnvalue = UnpackCounterBytes('<H', self.stream)
        if returnvalue == []:
            return result
        urlansi, streamRemainder = returnvalue
        result.append('URL(ansi): %s' % urlansi.decode('latin'))

        returnvalue = UnpackCounterBytes('<H', streamRemainder)
        if returnvalue == []:
            return result
        itemansi, streamRemainder = returnvalue
        result.append('Item(ansi): %s' % itemansi.decode('latin'))

        returnvalue = Unpack('<I', streamRemainder)
        if returnvalue == []:
            return result
        unknown, streamRemainder = returnvalue
        result.append('Unknown: %08x' % unknown)

        returnvalue = UnpackCounterBytes2('<H', streamRemainder)
        if returnvalue == []:
            return result
        urlunicode, streamRemainder = returnvalue
        result.append('URL(unicode): %s' % urlunicode.decode('utf16'))

        returnvalue = UnpackCounterBytes2('<H', streamRemainder)
        if returnvalue == []:
            return result
        itemunicode, streamRemainder = returnvalue
        result.append('Item(unicode): %s' % itemunicode.decode('utf16'))

        if len(streamRemainder) > 0:
            result.append('extra bytes: %d' % len(streamRemainder))

        return result

    #https://docs.microsoft.com/en-us/openspecs/office_file_formats/ms-doc/13ba10a8-d8b2-433b-bf3b-ec238dc8f9ce
    def Analyze3ObjInfo(self):
        result = []
        dCF = {
            0x0001: 'Rich Text Format',
            0x0002: 'Text format',
            0x0003: 'Metafile or Enhanced Metafile, depending on ODTPersist2.fStoredAsEMF',
            0x0004: 'Bitmap',
            0x0005: 'Device Independent Bitmap',
            0x000A: 'HTML format',
            0x0014: 'Unicode text format',
        }
        dODTPersist1 = {
            'A': 'reserved1 (1 bit): Undefined and MUST be ignored.',
            'B': 'fDefHandler (1 bit): If this bit is 1, then the application MUST assume that this OLE object’s class identifier (CLSID) is {00020907-0000-0000-C000-000000000046}.',
            'C': 'reserved2 (1 bit): Undefined and MUST be ignored.',
            'D': 'reserved3 (1 bit): Undefined and MUST be ignored.',
            'E': 'fLink (1 bit): A bit that specifies whether this OLE object is a link.',
            'F': 'reserved4 (1 bit): Undefined and MUST be ignored.',
            'G': 'fIcon (1 bit): A bit that specifies whether this OLE object is being represented by an icon.',
            'H': 'fIsOle1 (1 bit): A bit that specifies whether this OLE object is only compatible with OLE 1. If this bit is zero, then the object is compatible with OLE 2.',
            'I': 'fManual (1 bit): A bit that specifies whether the user has requested that this OLE object only be updated in response to a user action. If fManual is zero, then the user has requested that this OLE object update automatically. If fLink is zero, then fManual is undefined and MUST be ignored.',
            'J': 'fRecomposeOnResize (1 bit): A bit that specifies whether this OLE object has requested to be notified when it is resized by its container.',
            'K': 'reserved5 (1 bit): MUST be zero and MUST be ignored.',
            'L': 'reserved6 (1 bit): MUST be zero and MUST be ignored.',
            'M': 'fOCX (1 bit): A bit that specifies whether this object is an OLE control.',
            'N': 'fStream (1 bit): If fOCX is zero, then this bit MUST be zero. If fOCX is 1, then fStream is a bit that specifies whether this OLE control stores its data in a single stream instead of a storage. If fStream is 1, then the data for the OLE control is in a stream called "\003OCXDATA" where \003 is the character with value 0x0003, not the string literal "\003".',
            'O': 'reserved7 (1 bit): Undefined and MUST be ignored.',
            'P': 'fViewObject (1 bit): A bit that specifies whether this OLE object supports the IViewObject interface as described in [MSDOCS-IViewObject].',
        }
        dODTPersist2 = {
            'A': 'fEMF (1 bit): A bit that specifies that the presentation of this OLE object in the document is in the Enhanced Metafile format. This is different from fStoredAsEMF in the case of an object being represented as an icon. For icons, the icon can be an Enhanced Metafile even if the OLE object does not support the Enhanced Metafile format.',
            'B': 'reserved1 (1 bit): MUST be zero and MUST be ignored.',
            'C': 'fQueriedEMF (1 bit): A bit that specifies whether the application that saved this Word Binary file had queried this OLE object to determine whether it supported the Enhanced Metafile format.',
            'D': 'fStoredAsEMF (1 bit): A bit that specifies that this OLE object supports the Enhanced Metafile format.',
            'E': 'reserved2 (1 bit): Undefined and MUST be ignored.',
            'F': 'reserved3 (1 bit): Undefined and MUST be ignored.',
        }

        returnvalue = Unpack('<H', self.stream)
        if returnvalue == []:
            return result
        odtpersist1, streamRemainder = returnvalue
        result.append('odtpersist1: %s' % bin(odtpersist1))
        result.extend(Indent('  ', InterpretBits(odtpersist1, dODTPersist1)))

        returnvalue = Unpack('<H', streamRemainder)
        if returnvalue == []:
            return result
        cf, streamRemainder = returnvalue
        result.append('cf: %d (%s)' % (cf, dCF.get(cf, '<UNKNOWN>')))

        returnvalue = Unpack('<H', streamRemainder)
        if returnvalue == []:
            return result
        odtpersist2, streamRemainder = returnvalue
        result.append('odtpersist2: %s' % bin(odtpersist2))
        result.extend(Indent('  ', InterpretBits(odtpersist2, dODTPersist2)))

        if len(streamRemainder) > 0:
            result.append('extra bytes: %d' % len(streamRemainder))

        return result

    def Analyze1CompObj(self):
        result = []

        returnvalue = Unpack('<I', self.stream)
        if returnvalue == []:
            return result
        reserved1, streamRemainder = returnvalue
        result.append('Reserved1: %08x' % reserved1)

        returnvalue = Unpack('<I', streamRemainder)
        if returnvalue == []:
            return result
        version, streamRemainder = returnvalue
        result.append('Version: %08x' % version)

        returnvalue = UnpackBytes(20, streamRemainder)
        if returnvalue == []:
            return result
        reserved2, streamRemainder = returnvalue
        result.append('Reserved2')
        resultCLSIDs = FindCLSIDs(reserved2)
        if len(resultCLSIDs) > 0:
            result.append('  Contains known CLSIDs:')
            result.extend(Indent('    ', resultCLSIDs))

        returnvalue = UnpackCounterBytes('<I', streamRemainder)
        if returnvalue == []:
            return result
        ansiUserType, streamRemainder = returnvalue
        result.append('AnsiUserType: %s' % ansiUserType.decode('latin').rstrip('\x00'))

        returnvalue = UnpackClipboardFormatOrString(streamRemainder)
        if returnvalue == []:
            return result
        temp1, temp2, streamRemainder = returnvalue
        if temp1 == 0:
            result.append('ClipboardFormatOrAnsiString not present')
        elif temp1 in [0xFFFFFFFE, 0xFFFFFFFF]:
            result.append('ClipboardFormatOrAnsiString: %08x' % temp2)
        else:
            result.append('ClipboardFormatOrAnsiString: %s' % temp2.decode('latin').rstrip('\x00'))

        returnvalue = Unpack('<I', streamRemainder)
        if returnvalue == []:
            return result
        reserved1, streamRemainder = returnvalue
        if reserved1 == 0 or reserved1 > 0x28:
            result.append('Reserved1: 0x%08x (ignoring remainder)' % reserved1)
        else:
            returnvalue = UnpackBytes(reserved1, streamRemainder)
            if returnvalue == []:
                return result
            stringAnsi, streamRemainder = returnvalue
            result.append('Reserved1: %s' % stringAnsi.decode('latin').rstrip('\x00'))

            returnvalue = Unpack('<I', streamRemainder)
            if returnvalue == []:
                return result
            unicodeMarker, streamRemainder = returnvalue
            result.append('UnicodeMarker: 0x%08x' % unicodeMarker)
            if unicodeMarker == 0x71B239F4:
                returnvalue = UnpackCounterBytes('<I', streamRemainder)
                if returnvalue == []:
                    return result
                unicodeUserType, streamRemainder = returnvalue
                result.append('UnicodeUserType: %s' % unicodeUserType.decode('utf16').rstrip('\x00'))

                returnvalue = UnpackClipboardFormatOrString(streamRemainder)
                if returnvalue == []:
                    return result
                temp1, temp2, streamRemainder = returnvalue
                if temp1 == 0:
                    result.append('ClipboardFormatOrUnicodeString not present')
                elif temp1 in [0xFFFFFFFE, 0xFFFFFFFF]:
                    result.append('ClipboardFormatOrUnicodeString: %08x' % temp2)
                else:
                    result.append('ClipboardFormatOrUnicodeString: %s' % temp2.decode('utf16').rstrip('\x00'))

                returnvalue = UnpackCounterBytes('<I', streamRemainder)
                if returnvalue == []:
                    return result
                reserved2, streamRemainder = returnvalue
                result.append('Reserved2: %s' % reserved2.decode('utf16').rstrip('\x00'))

        if len(streamRemainder) > 0:
            result.append('extra bytes: %d' % len(streamRemainder))
            result.append('  %s' % repr(streamRemainder))

        return result

    def Analyze(self):
        result = []
        self.ran = True
        stream = self.stream

        if self.streamname[-1] == '\x01Ole':
            result = self.Analyze1Ole()
        elif self.streamname[-1] == '\x03LinkInfo':
            result = self.Analyze3LinkInfo()
        elif self.streamname[-1] == '\x03ObjInfo':
            result = self.Analyze3ObjInfo()
        elif self.streamname[-1] == '\x01CompObj':
            result = self.Analyze1CompObj()
        else:
            if KNOWN_CLSIDS == {}:
                result.append('<oletools missing>')
            for clsid, desc in KNOWN_CLSIDS.items():
                for position in FindAll(stream, GUIDToBytes(clsid)):
                    result.append('0x%08x %s %s' % (position, clsid, desc))

            return sorted(result)

        return result

AddPlugin(cCLSID)
